@@ -3,10 +3,16 @@ import logging
 import os
 import socket
 import time
+from typing import Tuple, Optional
 
 from fsspec import AbstractFileSystem
 from fsspec.spec import AbstractBufferedFile
-import boto3
+
+from fsspec.utils import infer_storage_options
+from fsspec.utils import tokenize
+
+import botocore
+import botocore.session
 from botocore.client import Config
 from botocore.exceptions import ClientError, ParamValidationError, BotoCoreError
 
@@ -14,47 +20,22 @@ from s3fs.errors import translate_boto_error
 from s3fs.utils import ParamKwargsHelper
 
 logger = logging.getLogger('s3fs')
-handle = logging.StreamHandler()
-formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s '
-                              '- %(message)s')
-handle.setFormatter(formatter)
-logger.addHandler(handle)
+
 if "S3FS_LOGGING_LEVEL" in os.environ:
+    handle = logging.StreamHandler()
+    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s '
+                                  '- %(message)s')
+    handle.setFormatter(formatter)
+    logger.addHandler(handle)
     logger.setLevel(os.environ["S3FS_LOGGING_LEVEL"])
 
-try:
-    from boto3.s3.transfer import S3_RETRYABLE_ERRORS
-except ImportError:
-    S3_RETRYABLE_ERRORS = (
-        socket.timeout,
-    )
+S3_RETRYABLE_ERRORS = (
+    socket.timeout,
+)
 
 _VALID_FILE_MODES = {'r', 'w', 'a', 'rb', 'wb', 'ab'}
 
 
-def split_path(path):
-    """
-    Normalise S3 path string into bucket and key.
-
-    Parameters
-    ----------
-    path : string
-        Input path, like `s3://mybucket/path/to/file`
-
-    Examples
-    --------
-    >>> split_path("s3://mybucket/path/to/file")
-    ['mybucket', 'path/to/file']
-    """
-    if path.startswith('s3://'):
-        path = path[5:]
-    path = path.rstrip('/').lstrip('/')
-    if '/' not in path:
-        return path, ""
-    else:
-        return path.split('/', 1)
-
-
 key_acls = {'private', 'public-read', 'public-read-write',
             'authenticated-read', 'aws-exec-read', 'bucket-owner-read',
             'bucket-owner-full-control'}
@@ -62,6 +43,32 @@ buck_acls = {'private', 'public-read', 'public-read-write',
              'authenticated-read'}
 
 
+def version_id_kw(version_id):
+    """Helper to make versionId kwargs.
+
+    Not all boto3 methods accept a None / empty versionId so dictionary expansion solves
+    that problem.
+    """
+    if version_id:
+        return {'VersionId': version_id}
+    else:
+        return {}
+
+
+def _coalesce_version_id(*args):
+    """Helper to coalesce a list of version_ids down to one"""
+    version_ids = set(args)
+    if None in version_ids:
+        version_ids.remove(None)
+    if len(version_ids) > 1:
+        raise ValueError(
+            f"Cannot coalesce version_ids where more than one are defined, {version_ids}")
+    elif len(version_ids) == 0:
+        return None
+    else:
+        return version_ids.pop()
+
+
 class S3FileSystem(AbstractFileSystem):
     """
     Access S3 as if it were a file system.
@@ -70,7 +77,7 @@ class S3FileSystem(AbstractFileSystem):
     storage.
 
     Provide credentials either explicitly (``key=``, ``secret=``) or depend
-    on boto's credential methods. See boto3 documentation for more
+    on boto's credential methods. See botocore documentation for more
     information. If no credentials are available, use ``anon=True``.
 
     Parameters
@@ -90,7 +97,7 @@ class S3FileSystem(AbstractFileSystem):
         insecure
     s3_additional_kwargs : dict of parameters that are used when calling s3 api
         methods. Typically used for things like "ServerSideEncryption".
-    client_kwargs : dict of parameters for the boto3 client
+    client_kwargs : dict of parameters for the botocore client
     requester_pays : bool (False)
         If RequesterPays buckets are supported.
     default_block_size: int (None)
@@ -108,10 +115,14 @@ class S3FileSystem(AbstractFileSystem):
         user to have the necessary IAM permissions for dealing with versioned
         objects.
     config_kwargs : dict of parameters passed to ``botocore.client.Config``
-    kwargs : other parameters for boto3 session
+    kwargs : other parameters for core session
     session : botocore Session object to be used for all connections.
          This session will be used inplace of creating a new session inside S3FileSystem.
 
+    The following parameters are passed on to fsspec:
+
+    skip_instance_cache: to control reuse of instances
+    use_listings_cache, listings_expiry_time, max_paths: to control reuse of directory listings
 
     Examples
     --------
@@ -125,9 +136,10 @@ class S3FileSystem(AbstractFileSystem):
     """
     root_marker = ""
     connect_timeout = 5
+    retries = 5
     read_timeout = 15
     default_block_size = 5 * 2**20
-    protocol = 's3'
+    protocol = ['s3', 's3a']
     _extra_tokenize_attributes = ('default_block_size',)
 
     def __init__(self, anon=False, key=None, secret=None, token=None,
@@ -154,6 +166,9 @@ class S3FileSystem(AbstractFileSystem):
         self.secret = secret
         self.token = token
         self.kwargs = kwargs
+        super_kwargs = {k: kwargs.pop(k)
+                        for k in ['use_listings_cache', 'listings_expiry_time', 'max_paths']
+                        if k in kwargs}  # passed to fsspec superclass
 
         if client_kwargs is None:
             client_kwargs = {}
@@ -170,7 +185,7 @@ class S3FileSystem(AbstractFileSystem):
         self.use_ssl = use_ssl
         self.s3 = self.connect()
         self._kwargs_helper = ParamKwargsHelper(self.s3)
-        super().__init__()
+        super().__init__(**super_kwargs)
 
     def _filter_kwargs(self, s3_method, kwargs):
         return self._kwargs_helper.filter_dict(s3_method.__name__, kwargs)
@@ -192,6 +207,56 @@ class S3FileSystem(AbstractFileSystem):
         # filter all kwargs
         return self._filter_kwargs(method, additional_kwargs)
 
+    @staticmethod
+    def _get_kwargs_from_urls(urlpath):
+        """
+        When we have a urlpath that contains a ?versionId= query assume that we want to use version_aware mode for
+        the filesystem.
+        """
+        url_storage_opts = infer_storage_options(urlpath)
+        url_query = url_storage_opts.get("url_query")
+        out = {}
+        if url_query is not None:
+            from urllib.parse import parse_qs
+            parsed = parse_qs(url_query)
+            if 'versionId' in parsed:
+                out['version_aware'] = True
+        return out
+
+    def split_path(self, path) -> Tuple[str, str, Optional[str]]:
+        """
+        Normalise S3 path string into bucket and key.
+
+        Parameters
+        ----------
+        path : string
+            Input path, like `s3://mybucket/path/to/file`
+
+        Examples
+        --------
+        >>> split_path("s3://mybucket/path/to/file")
+        ['mybucket', 'path/to/file', None]
+
+        >>> split_path("s3://mybucket/path/to/versioned_file?versionId=some_version_id")
+        ['mybucket', 'path/to/versioned_file', 'some_version_id']
+        """
+        path = self._strip_protocol(path)
+        path = path.lstrip('/')
+        if '/' not in path:
+            return path, "", None
+        else:
+            bucket, keypart = path.split('/', 1)
+            key, _, version_id = keypart.partition('?versionId=')
+            return bucket, key, version_id if self.version_aware and version_id else None
+
+    def _prepare_config_kwargs(self):
+        config_kwargs = self.config_kwargs.copy()
+        if "connect_timeout" not in config_kwargs.keys():
+            config_kwargs['connect_timeout'] = self.connect_timeout
+        if "read_timeout" not in config_kwargs.keys():
+            config_kwargs['read_timeout'] = self.read_timeout
+        return config_kwargs
+
     def connect(self, refresh=True):
         """
         Establish S3 connection object.
@@ -210,24 +275,21 @@ class S3FileSystem(AbstractFileSystem):
             self.anon, self.key, self.secret, self.kwargs,
             self.client_kwargs, self.token, self.use_ssl)
 
-        if self.anon:
-            from botocore import UNSIGNED
-            conf = Config(connect_timeout=self.connect_timeout,
-                          read_timeout=self.read_timeout,
-                          signature_version=UNSIGNED, **self.config_kwargs)
-            if not self.passed_in_session:
-                self.session = boto3.Session(**self.kwargs)
-        else:
-            conf = Config(connect_timeout=self.connect_timeout,
-                          read_timeout=self.read_timeout,
-                          **self.config_kwargs)
-            if not self.passed_in_session:
-                self.session = boto3.Session(self.key, self.secret, self.token,
-                                             **self.kwargs)
+        if not self.passed_in_session:
+            self.session = botocore.session.Session(**self.kwargs)
 
         logger.debug("Setting up s3fs instance")
-        self.s3 = self.session.client('s3', config=conf, use_ssl=ssl,
-                                      **self.client_kwargs)
+
+        config_kwargs = self._prepare_config_kwargs()
+        if self.anon:
+            from botocore import UNSIGNED
+            conf = Config(signature_version=UNSIGNED, **config_kwargs)
+            self.s3 = self.session.create_client('s3', config=conf, use_ssl=ssl,
+                                        **self.client_kwargs)
+        else:
+            conf = Config(**config_kwargs)
+            self.s3 = self.session.create_client('s3', aws_access_key_id=self.key, aws_secret_access_key=self.secret, aws_session_token=self.token, config=conf, use_ssl=ssl,
+                                        **self.client_kwargs)
         return self.s3
 
     def get_delegated_s3pars(self, exp=3600):
@@ -250,7 +312,7 @@ class S3FileSystem(AbstractFileSystem):
                     'anon': False}
         if self.key is None or self.secret is None:  # automatic credentials
             return {'anon': False}
-        sts = self.session.client('sts')
+        sts = self.session.create_client('sts')
         cred = sts.get_session_token(DurationSeconds=exp)['Credentials']
         return {'key': cred['AccessKeyId'], 'secret': cred['SecretAccessKey'],
                 'token': cred['SessionToken'], 'anon': False}
@@ -315,10 +377,7 @@ class S3FileSystem(AbstractFileSystem):
                       autocommit=autocommit, requester_pays=requester_pays)
 
     def _lsdir(self, path, refresh=False, max_items=None):
-        if path.startswith('s3://'):
-            path = path[len('s3://'):]
-        path = path.rstrip('/')
-        bucket, prefix = split_path(path)
+        bucket, prefix, _ = self.split_path(path)
         prefix = prefix + '/' if prefix else ""
         if path not in self.dircache or refresh:
             try:
@@ -349,6 +408,7 @@ class S3FileSystem(AbstractFileSystem):
                 raise translate_boto_error(e)
 
             self.dircache[path] = files
+            return files
         return self.dircache[path]
 
     def mkdir(self, path, acl="", **kwargs):
@@ -401,6 +461,7 @@ class S3FileSystem(AbstractFileSystem):
                 f['name'] = f['Name']
                 del f['Name']
             self.dircache[''] = files
+            return files
         return self.dircache['']
 
     def _ls(self, path, refresh=False):
@@ -418,8 +479,7 @@ class S3FileSystem(AbstractFileSystem):
         refresh : bool (=False)
             if False, look in local cache for file details first
         """
-        if path.startswith('s3://'):
-            path = path[len('s3://'):]
+        path = self._strip_protocol(path)
         if path in ['', '/']:
             return self._lsbuckets(refresh)
         else:
@@ -429,7 +489,7 @@ class S3FileSystem(AbstractFileSystem):
         if path in ['', '/']:
             # the root always exists, even if anon
             return True
-        bucket, key = split_path(path)
+        bucket, key, version_id = self.split_path(path)
         if key:
             return super().exists(path)
         else:
@@ -441,7 +501,9 @@ class S3FileSystem(AbstractFileSystem):
 
     def touch(self, path, truncate=True, data=None, **kwargs):
         """Create empty file or truncate"""
-        bucket, key = split_path(path)
+        bucket, key, version_id = self.split_path(path)
+        if version_id:
+            raise ValueError("S3 does not support touching existing versions of files")
         if not truncate and self.exists(path):
             raise ValueError("S3 does not support touching existent files")
         try:
@@ -450,7 +512,7 @@ class S3FileSystem(AbstractFileSystem):
             raise translate_boto_error(ex)
         self.invalidate_cache(self._parent(path))
 
-    def info(self, path, version_id=None):
+    def info(self, path, version_id=None, refresh=False):
         if path in ['/', '']:
             return {'name': path, 'size': 0, 'type': 'directory'}
         kwargs = self.kwargs.copy()
@@ -458,19 +520,20 @@ class S3FileSystem(AbstractFileSystem):
             if not self.version_aware:
                 raise ValueError("version_id cannot be specified if the "
                                  "filesystem is not version aware")
-            kwargs['VersionId'] = version_id
-        if self.version_aware:
+        bucket, key, path_version_id = self.split_path(path)
+        version_id = _coalesce_version_id(path_version_id, version_id)
+        if self.version_aware or (key and self._ls_from_cache(path) is None) or refresh:
             try:
-                bucket, key = split_path(path)
                 out = self._call_s3(self.s3.head_object, kwargs, Bucket=bucket,
-                                    Key=key, **self.req_kw)
+                                    Key=key, **version_id_kw(version_id), **self.req_kw)
                 return {
                     'ETag': out['ETag'],
                     'Key': '/'.join([bucket, key]),
                     'LastModified': out['LastModified'],
                     'Size': out['ContentLength'],
                     'size': out['ContentLength'],
-                    'path': '/'.join([bucket, key]),
+                    'name': '/'.join([bucket, key]),
+                    'type': 'file',
                     'StorageClass': "STANDARD",
                     'VersionId': out.get('VersionId')
                 }
@@ -478,12 +541,37 @@ class S3FileSystem(AbstractFileSystem):
                 ee = translate_boto_error(e)
                 # This could have failed since the thing we are looking for is a prefix.
                 if isinstance(ee, FileNotFoundError):
-                    return super().info(path)
+                    return super(S3FileSystem, self).info(path)
                 else:
                     raise ee
             except ParamValidationError as e:
                 raise ValueError('Failed to head path %r: %s' % (path, e))
         return super().info(path)
+    
+    def checksum(self, path, refresh=False):
+        """
+        Unique value for current version of file
+
+        If the checksum is the same from one moment to another, the contents
+        are guaranteed to be the same. If the checksum changes, the contents
+        *might* have changed.
+
+        Parameters
+        ----------
+        path : string/bytes
+            path of file to get checksum for
+        refresh : bool (=False)
+            if False, look in local cache for file details first
+        
+        """
+
+        info = self.info(path, refresh=refresh)
+        
+        if info["type"] != 'directory':
+            return int(info["ETag"].strip('"'), 16)
+        else:
+            return int(tokenize(info), 16)
+        
 
     def isdir(self, path):
         path = self._strip_protocol(path).strip("/")
@@ -540,7 +628,7 @@ class S3FileSystem(AbstractFileSystem):
         if not self.version_aware:
             raise ValueError("version specific functionality is disabled for "
                              "non-version aware filesystems")
-        bucket, key = split_path(path)
+        bucket, key, _ = self.split_path(path)
         kwargs = {}
         out = {'IsTruncated': True}
         versions = []
@@ -565,13 +653,13 @@ class S3FileSystem(AbstractFileSystem):
         refresh : bool (=False)
             if False, look in local cache for file metadata first
         """
-        bucket, key = split_path(path)
-
+        bucket, key, version_id = self.split_path(path)
         if refresh or path not in self._metadata_cache:
             response = self._call_s3(self.s3.head_object,
                                      kwargs,
                                      Bucket=bucket,
                                      Key=key,
+                                     **version_id_kw(version_id),
                                      **self.req_kw)
             self._metadata_cache[path] = response['Metadata']
 
@@ -584,9 +672,9 @@ class S3FileSystem(AbstractFileSystem):
         -------
         {str: str}
         """
-        bucket, key = split_path(path)
+        bucket, key, version_id = self.split_path(path)
         response = self._call_s3(self.s3.get_object_tagging,
-                                 Bucket=bucket, Key=key)
+                                 Bucket=bucket, Key=key, **version_id_kw(version_id))
         return {v['Key']: v['Value'] for v in response['TagSet']}
 
     def put_tags(self, path, tags, mode='o'):
@@ -610,7 +698,7 @@ class S3FileSystem(AbstractFileSystem):
             'm': Will merge in new tags with existing tags.  Incurs two remote
             calls.
         """
-        bucket, key = split_path(path)
+        bucket, key, version_id = self.split_path(path)
 
         if mode == 'm':
             existing_tags = self.get_tags(path=path)
@@ -624,7 +712,7 @@ class S3FileSystem(AbstractFileSystem):
 
         tag = {'TagSet': new_tags}
         self._call_s3(self.s3.put_object_tagging,
-                      Bucket=bucket, Key=key, Tagging=tag)
+                      Bucket=bucket, Key=key, Tagging=tag, **version_id_kw(version_id))
 
     def getxattr(self, path, attr_name, **kwargs):
         """ Get an attribute from the metadata.
@@ -667,7 +755,7 @@ class S3FileSystem(AbstractFileSystem):
         http://docs.aws.amazon.com/AmazonS3/latest/dev/UsingMetadata.html#object-metadata
         """
 
-        bucket, key = split_path(path)
+        bucket, key, version_id = self.split_path(path)
         metadata = self.metadata(path)
         metadata.update(**kw_args)
         copy_kwargs = copy_kwargs or {}
@@ -677,10 +765,14 @@ class S3FileSystem(AbstractFileSystem):
             if kw_args[kw_key] is None:
                 metadata.pop(kw_key, None)
 
+        src = {'Bucket': bucket, 'Key': key}
+        if version_id:
+            src['VersionId'] = version_id
+
         self._call_s3(
             self.s3.copy_object,
             copy_kwargs,
-            CopySource="{}/{}".format(bucket, key),
+            CopySource=src,
             Bucket=bucket,
             Key=key,
             Metadata=metadata,
@@ -702,12 +794,12 @@ class S3FileSystem(AbstractFileSystem):
         acl : string
             the value of ACL to apply
         """
-        bucket, key = split_path(path)
+        bucket, key, version_id = self.split_path(path)
         if key:
             if acl not in key_acls:
                 raise ValueError('ACL not in %s', key_acls)
             self._call_s3(self.s3.put_object_acl,
-                          kwargs, Bucket=bucket, Key=key, ACL=acl)
+                          kwargs, Bucket=bucket, Key=key, ACL=acl, **version_id_kw(version_id))
         else:
             if acl not in buck_acls:
                 raise ValueError('ACL not in %s', buck_acls)
@@ -724,10 +816,10 @@ class S3FileSystem(AbstractFileSystem):
         expires : int
             the number of seconds this signature will be good for.
         """
-        bucket, key = split_path(path)
+        bucket, key, version_id = self.split_path(path)
         return self.s3.generate_presigned_url(
             ClientMethod='get_object',
-            Params=dict(Bucket=bucket, Key=key, **kwargs),
+            Params=dict(Bucket=bucket, Key=key, **version_id_kw(version_id), **kwargs),
             ExpiresIn=expires)
 
     def merge(self, path, filelist, **kwargs):
@@ -743,13 +835,16 @@ class S3FileSystem(AbstractFileSystem):
         filelist : list of str
             The paths, in order, to assemble into the final file.
         """
-        bucket, key = split_path(path)
+        bucket, key, version_id = self.split_path(path)
+        if version_id:
+            raise ValueError("Cannot write to an explicit versioned file!")
         mpu = self._call_s3(
             self.s3.create_multipart_upload,
             kwargs,
             Bucket=bucket,
             Key=key
         )
+        # TODO: Make this support versions?
         out = [self._call_s3(
             self.s3.upload_part_copy,
             kwargs,
@@ -765,45 +860,85 @@ class S3FileSystem(AbstractFileSystem):
         self.invalidate_cache(path)
 
     def copy_basic(self, path1, path2, **kwargs):
-        """ Copy file between locations on S3 """
-        buc1, key1 = split_path(path1)
-        buc2, key2 = split_path(path2)
+        """ Copy file between locations on S3
+
+        Not allowed where the origin is >5GB - use copy_managed
+        """
+        buc1, key1, ver1 = self.split_path(path1)
+        buc2, key2, ver2 = self.split_path(path2)
+        if ver2:
+            raise ValueError("Cannot copy to a versioned file!")
         try:
+            copy_src = {'Bucket': buc1, 'Key': key1}
+            if ver1:
+                copy_src['VersionId'] = ver1
             self._call_s3(
                 self.s3.copy_object,
                 kwargs,
-                Bucket=buc2, Key=key2, CopySource='/'.join([buc1, key1])
+                Bucket=buc2, Key=key2, CopySource=copy_src
             )
         except ClientError as e:
             raise translate_boto_error(e)
         except ParamValidationError as e:
             raise ValueError('Copy failed (%r -> %r): %s' % (path1, path2, e))
+        self.invalidate_cache(path2)
 
-    def copy_managed(self, path1, path2, **kwargs):
-        buc1, key1 = split_path(path1)
-        buc2, key2 = split_path(path2)
-        copy_source = {
-            'Bucket': buc1,
-            'Key': key1
-        }
-        try:
-            self.s3.copy(
-                CopySource=copy_source,
-                Bucket=buc2,
-                Key=key2,
-                ExtraArgs=self._get_s3_method_kwargs(
-                    self.s3.copy_object,
-                    kwargs
-                )
-            )
-        except ClientError as e:
-            raise translate_boto_error(e)
-        except ParamValidationError as e:
-            raise ValueError('Copy failed (%r -> %r): %s' % (path1, path2, e))
+    def copy_managed(self, path1, path2, block=5 * 2**30, **kwargs):
+        """Copy file between locations on S3 as multi-part
+
+        block: int
+            The size of the pieces, must be larger than 5MB and at most 5GB.
+            Smaller blocks mean more calls, only useful for testing.
+        """
+        if block < 5*2**20 or block > 5*2**30:
+            raise ValueError('Copy block size must be 5MB<=block<=5GB')
+        size = self.info(path1)['size']
+        bucket, key, version = self.split_path(path2)
+        mpu = self._call_s3(
+            self.s3.create_multipart_upload,
+            Bucket=bucket, Key=key, **kwargs)
+        parts = []
+        for i, offset in enumerate(range(0, size + 1, block)):
+            for attempt in range(self.retries + 1):
+                try:
+                    brange = "bytes=%i-%i" % (
+                            offset, min(offset + block - 1, size))
+                    out = self._call_s3(
+                        self.s3.upload_part_copy,
+                        Bucket=bucket,
+                        Key=key,
+                        PartNumber=i + 1,
+                        UploadId=mpu['UploadId'],
+                        CopySource=path1,
+                        CopySourceRange=brange)
+                    break
+                except S3_RETRYABLE_ERRORS as exc:
+                    if attempt < self.retries:
+                        logger.debug('Exception %r on S3 write, retrying', exc,
+                                     exc_info=True)
+                    time.sleep(1.7 ** attempt * 0.1)
+                except Exception as exc:
+                    raise IOError('Write failed: %r' % exc)
+            parts.append({'PartNumber': i + 1,
+                          'ETag': out['CopyPartResult']['ETag']})
+        self._call_s3(
+            self.s3.complete_multipart_upload,
+            Bucket=bucket,
+            Key=key,
+            UploadId=mpu['UploadId'],
+            MultipartUpload={'Parts': parts})
+        self.invalidate_cache(path2)
 
     def copy(self, path1, path2, **kwargs):
-        self.copy_managed(path1, path2, **kwargs)
-        self.invalidate_cache(path2)
+        gb5 = 5 * 2**30
+        path1 = self._strip_protocol(path1)
+        size = self.info(path1)['size']
+        if size <= gb5:
+            # simple copy allowed for <5GB
+            self.copy_basic(path1, path2, **kwargs)
+        else:
+            # serial multipart copy
+            self.copy_managed(path1, path2, **kwargs)
 
     def bulk_delete(self, pathlist, **kwargs):
         """
@@ -816,7 +951,7 @@ class S3FileSystem(AbstractFileSystem):
         """
         if not pathlist:
             return
-        buckets = {split_path(path)[0] for path in pathlist}
+        buckets = {self.split_path(path)[0] for path in pathlist}
         if len(buckets) > 1:
             raise ValueError("Bulk delete files should refer to only one "
                              "bucket")
@@ -825,7 +960,7 @@ class S3FileSystem(AbstractFileSystem):
             for i in range((len(pathlist) // 1000) + 1):
                 self.bulk_delete(pathlist[i * 1000:(i + 1) * 1000])
             return
-        delete_keys = {'Objects': [{'Key': split_path(path)[1]} for path
+        delete_keys = {'Objects': [{'Key': self.split_path(path)[1]} for path
                                    in pathlist]}
         for path in pathlist:
             self.invalidate_cache(self._parent(path))
@@ -849,7 +984,8 @@ class S3FileSystem(AbstractFileSystem):
             Whether to remove also all entries below, i.e., which are returned
             by `walk()`.
         """
-        bucket, key = split_path(path)
+        # TODO: explicit version deletes?
+        bucket, key, _ = self.split_path(path)
         if recursive:
             files = self.find(path, maxdepth=None)
             if key and not files:
@@ -884,10 +1020,12 @@ class S3FileSystem(AbstractFileSystem):
         else:
             path = self._strip_protocol(path)
             self.dircache.pop(path, None)
-            self.dircache.pop(self._parent(path), None)
+            while path:
+                self.dircache.pop(path, None)
+                path = self._parent(path)
 
     def walk(self, path, maxdepth=None, **kwargs):
-        if path in ['', '*', 's3://']:
+        if path in ['', '*'] + [f'{p}://' for p in self.protocol]:
             raise ValueError('Cannot crawl all of S3')
         return super().walk(path, maxdepth=maxdepth, **kwargs)
 
@@ -899,7 +1037,7 @@ class S3File(AbstractBufferedFile):
     Parameters
     ----------
     s3 : S3FileSystem
-        boto3 connection
+        botocore connection
     path : string
         S3 bucket/key to access
     mode : str
@@ -939,12 +1077,12 @@ class S3File(AbstractBufferedFile):
     def __init__(self, s3, path, mode='rb', block_size=5 * 2 ** 20, acl="",
                  version_id=None, fill_cache=True, s3_additional_kwargs=None,
                  autocommit=True, cache_type='bytes', requester_pays=False):
-        bucket, key = split_path(path)
+        bucket, key, path_version_id = s3.split_path(path)
         if not key:
             raise ValueError('Attempt to open non key-like path: %s' % path)
         self.bucket = bucket
         self.key = key
-        self.version_id = version_id
+        self.version_id = _coalesce_version_id(version_id, path_version_id)
         self.acl = acl
         if self.acl and self.acl not in key_acls:
             raise ValueError('ACL not in %s', key_acls)
@@ -991,7 +1129,7 @@ class S3File(AbstractBufferedFile):
                                 **kwargs)
 
     def _initiate_upload(self):
-        if not self.autocommit and not self.append_block and self.tell() < self.blocksize:
+        if self.autocommit and not self.append_block and self.tell() < self.blocksize:
             # only happens when closing small file, use on-shot PUT
             return
         logger.debug("Initiate upload for %s" % self)
@@ -1060,11 +1198,11 @@ class S3File(AbstractBufferedFile):
         return _fetch_range(self.fs.s3, self.bucket, self.key, self.version_id, start, end, req_kw=self.req_kw)
 
     def _upload_chunk(self, final=False):
-        bucket, key = split_path(self.path)
+        bucket, key, _ = self.fs.split_path(self.path)
         logger.debug("Upload for %s, final=%s, loc=%s, buffer loc=%s" % (
             self, final, self.loc, self.buffer.tell()
         ))
-        if not self.autocommit and not self.append_block and final and self.tell() < self.blocksize:
+        if self.autocommit and not self.append_block and final and self.tell() < self.blocksize:
             # only happens when closing small file, use on-shot PUT
             data1 = False
         else:
@@ -1124,10 +1262,12 @@ class S3File(AbstractBufferedFile):
                 logger.debug("One-shot upload of %s" % self)
                 self.buffer.seek(0)
                 data = self.buffer.read()
-                self._call_s3(
+                write_result = self._call_s3(
                     self.fs.s3.put_object,
                     Key=self.key, Bucket=self.bucket, Body=data, **self.kwargs
                 )
+                if self.fs.version_aware:
+                    self.version_id = write_result.get('VersionId')
             else:
                 raise RuntimeError
         else:
@@ -1188,13 +1328,10 @@ def _fetch_range(client, bucket, key, version_id, start, end, max_attempts=10,
     logger.debug("Fetch: %s/%s, %s-%s", bucket, key, start, end)
     for i in range(max_attempts):
         try:
-            if version_id is not None:
-                kwargs = dict({'VersionId': version_id}, **req_kw)
-            else:
-                kwargs = req_kw
             resp = client.get_object(Bucket=bucket, Key=key,
                                      Range='bytes=%i-%i' % (start, end - 1),
-                                     **kwargs)
+                                     **version_id_kw(version_id),
+                                     **req_kw)
             return resp['Body'].read()
         except S3_RETRYABLE_ERRORS as e:
             logger.debug('Exception %r on S3 download, retrying', e,
