@@ -11,7 +11,7 @@ import fsspec.core
 from s3fs.core import S3FileSystem
 from s3fs.utils import ignoring, SSEParams
 import moto
-import boto3
+import botocore
 from unittest import mock
 from botocore.exceptions import NoCredentialsError
 
@@ -47,23 +47,26 @@ b = test_bucket_name + '/tmp/test/b'
 c = test_bucket_name + '/tmp/test/c'
 d = test_bucket_name + '/tmp/test/d'
 
-
 @pytest.yield_fixture
 def s3():
     # writable local S3 system
     with moto.mock_s3():
-        import boto3
-        client = boto3.client('s3')
+        from botocore.session import Session
+        session = Session()
+        client = session.create_client('s3')
         client.create_bucket(Bucket=test_bucket_name, ACL='public-read')
 
-        bucket = client.create_bucket(
+        client.create_bucket(
             Bucket=versioned_bucket_name, ACL='public-read')
-        bucket_versioning = boto3.resource(
-            's3').BucketVersioning(versioned_bucket_name)
-        bucket_versioning.enable()
+        client.put_bucket_versioning(
+            Bucket=versioned_bucket_name,
+            VersioningConfiguration={
+                'Status': 'Enabled'
+            }
+        )
 
         # initialize secure bucket
-        bucket = client.create_bucket(
+        client.create_bucket(
             Bucket=secure_bucket_name, ACL='public-read')
         policy = json.dumps({
             "Version": "2012-10-17",
@@ -165,6 +168,23 @@ def test_config_kwargs():
     assert s3.connect(refresh=True).meta.config.signature_version == 's3v4'
 
 
+def test_config_kwargs_class_attributes_default():
+    s3 = S3FileSystem()
+    assert s3.connect(refresh=True).meta.config.connect_timeout == 5
+    assert s3.connect(refresh=True).meta.config.read_timeout == 15
+
+
+def test_config_kwargs_class_attributes_override():
+    s3 = S3FileSystem(
+        config_kwargs={
+            "connect_timeout": 60,
+            "read_timeout": 120,
+        }
+    )
+    assert s3.connect(refresh=True).meta.config.connect_timeout == 60
+    assert s3.connect(refresh=True).meta.config.read_timeout == 120
+
+
 def test_idempotent_connect(s3):
     con1 = s3.connect()
     con2 = s3.connect(refresh=False)
@@ -184,7 +204,11 @@ def test_multiple_objects(s3):
 def test_info(s3):
     s3.touch(a)
     s3.touch(b)
-    assert s3.info(a) == s3.ls(a, detail=True)[0]
+    info = s3.info(a)
+    linfo = s3.ls(a, detail=True)[0]
+    assert abs(info.pop('LastModified') - linfo.pop('LastModified')).seconds < 1
+    info.pop('VersionId')
+    assert info == linfo
     parent = a.rsplit('/', 1)[0]
     s3.invalidate_cache()  # remove full path from the cache
     s3.ls(parent)  # fill the cache with parent dir
@@ -200,6 +224,45 @@ def test_info(s3):
         s3.info(new_parent)
 
 
+def test_checksum(s3):
+    bucket = test_bucket_name
+    d = "checksum"
+    prefix = d+"/e"
+    o1 = prefix + "1"
+    o2 = prefix + "2"
+    path1 = bucket + "/" + o1
+    path2 = bucket + "/" + o2
+
+    client=s3.s3
+
+    # init client and files
+    client.put_object(Bucket=bucket, Key=o1, Body="")
+    client.put_object(Bucket=bucket, Key=o2, Body="")
+
+    # change one file, using cache
+    client.put_object(Bucket=bucket, Key=o1, Body="foo")
+    checksum = s3.checksum(path1)
+    s3.ls(path1) # force caching
+    client.put_object(Bucket=bucket, Key=o1, Body="bar")
+    # refresh == False => checksum doesn't change
+    assert checksum == s3.checksum(path1)
+
+    # change one file, without cache
+    client.put_object(Bucket=bucket, Key=o1, Body="foo")
+    checksum = s3.checksum(path1, refresh=True)
+    s3.ls(path1) # force caching
+    client.put_object(Bucket=bucket, Key=o1, Body="bar")
+    # refresh == True => checksum changes
+    assert checksum != s3.checksum(path1, refresh=True)
+
+
+    # Test for nonexistent file
+    client.put_object(Bucket=bucket, Key=o1, Body="bar")
+    s3.ls(path1) # force caching
+    client.delete_object(Bucket=bucket, Key=o1)
+    with pytest.raises(FileNotFoundError):
+        checksum = s3.checksum(o1, refresh=True)
+        
 test_xattr_sample_metadata = {'test_xattr': '1'}
 
 
@@ -598,6 +661,19 @@ def test_copy(s3):
     assert s3.cat(fn) == s3.cat(fn + '2')
 
 
+def test_copy_managed(s3):
+    data = b'abc' * 12*2**20
+    fn = test_bucket_name + '/test/biggerfile'
+    with s3.open(fn, 'wb') as f:
+        f.write(data)
+    s3.copy_managed(fn, fn + '2', block=5 * 2 ** 20)
+    assert s3.cat(fn) == s3.cat(fn + '2')
+    with pytest.raises(ValueError):
+        s3.copy_managed(fn, fn + '3', block=4 * 2 ** 20)
+    with pytest.raises(ValueError):
+        s3.copy_managed(fn, fn + '3', block=6 * 2 ** 30)
+
+
 def test_move(s3):
     fn = test_bucket_name + '/test/accounts.1.json'
     data = s3.cat(fn)
@@ -1083,6 +1159,7 @@ def test_versions(s3):
         fo.write(b'1')
     with s3.open(versioned_file, 'wb') as fo:
         fo.write(b'2')
+    assert s3.isfile(versioned_file)
     versions = s3.object_version_info(versioned_file)
     version_ids = [version['VersionId'] for version in versions]
     assert len(version_ids) == 2
@@ -1108,6 +1185,47 @@ def test_list_versions_many(s3):
     assert len(versions) == 1200
 
 
+def test_fsspec_versions_multiple(s3):
+    """Test that the standard fsspec.core.get_fs_token_paths behaves as expected for versionId urls"""
+    s3 = S3FileSystem(anon=False, version_aware=True)
+    versioned_file = versioned_bucket_name + '/versioned_file3'
+    version_lookup = {}
+    for i in range(20):
+        contents = str(i).encode()
+        with s3.open(versioned_file, 'wb') as fo:
+            fo.write(contents)
+        version_lookup[fo.version_id] = contents
+    urls = [f"s3://{versioned_file}?versionId={version}" for version in version_lookup.keys()]
+    fs, token, paths = fsspec.core.get_fs_token_paths(urls)
+    assert isinstance(fs, S3FileSystem)
+    assert fs.version_aware
+    for path in paths:
+        with fs.open(path, 'rb') as fo:
+            contents = fo.read()
+            assert contents == version_lookup[fo.version_id]
+
+
+def test_versioned_file_fullpath(s3):
+    versioned_file = versioned_bucket_name + '/versioned_file_fullpath'
+    s3 = S3FileSystem(anon=False, version_aware=True)
+    with s3.open(versioned_file, 'wb') as fo:
+        fo.write(b'1')
+    # moto doesn't correctly return a versionId for a multipart upload. So we resort to this.
+    # version_id = fo.version_id
+    versions = s3.object_version_info(versioned_file)
+    version_ids = [version['VersionId'] for version in versions]
+    version_id = version_ids[0]
+
+    with s3.open(versioned_file, 'wb') as fo:
+        fo.write(b'2')
+
+    file_with_version = "{}?versionId={}".format(versioned_file, version_id)
+
+    with s3.open(file_with_version, 'rb') as fo:
+        assert fo.version_id == version_id
+        assert fo.read() == b'1'
+
+
 def test_versions_unaware(s3):
     versioned_file = versioned_bucket_name + '/versioned_file3'
     s3 = S3FileSystem(anon=False, version_aware=False)
@@ -1205,7 +1323,7 @@ def test_change_defaults_only_subsequent():
 
 
 def test_passed_in_session_set_correctly(s3):
-    session = boto3.session.Session()
+    session = botocore.session.Session()
     s3 = S3FileSystem(session=session)
     assert s3.passed_in_session is session
     client = s3.connect()
@@ -1213,7 +1331,7 @@ def test_passed_in_session_set_correctly(s3):
 
 
 def test_without_passed_in_session_set_unique(s3):
-    session = boto3.session.Session()
+    session = botocore.session.Session()
     s3 = S3FileSystem()
     assert s3.passed_in_session is None
     client = s3.connect()
@@ -1228,9 +1346,9 @@ def test_pickle_without_passed_in_session(s3):
 
 def test_pickle_with_passed_in_session(s3):
     import pickle
-    session = boto3.session.Session()
+    session = botocore.session.Session()
     s3 = S3FileSystem(session=session)
-    with pytest.raises((AttributeError, NotImplementedError, TypeError)):
+    with pytest.raises((AttributeError, NotImplementedError, TypeError, pickle.PicklingError)):
         pickle.dumps(s3)
 
 
@@ -1276,9 +1394,8 @@ def test_autocommit(s3):
 def test_autocommit_mpu(s3):
     """When not autocommitting we always want to use multipart uploads"""
     path = test_bucket_name + '/auto_commit_with_mpu'
-    with s3.open(path, 'wb', autocommit=True) as fo:
+    with s3.open(path, 'wb', autocommit=False) as fo:
         fo.write(b'1')
-    # fo.flush()
     assert fo.mpu is not None
     assert len(fo.parts) == 1
 
